/********************************************************************/
/*                              Poly                                */
/*                                                                  */
/*   Luc Maranget, projet Moscova, Inria Rocquencourt               */
/*                                                                  */
/* Copyright 2005 Institut National de Recherche en Informatique et */
/* en Automatique.  All rights reserved.  This file is distributed  */
/* under the terms of the Q Public License version 1.0.             */ 
/********************************************************************/



/*
  Enumeration des polyominos following:

  Redelmeier, D. H. 'Counting Polyominoes: Yet Another Attack'.
  Discrete Math. 36, 191-203, 1981. 

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

typedef enum statut {libre, voisin, plein} statut;

typedef struct point {
  statut stat;
  struct point *next;
  struct point *voisins[5];
} point;

/* damier, comptes */

typedef unsigned long compteur;
#define INCR(x) ++(x)

void print_count(FILE *fp,compteur c)
{
  fprintf(fp,"%16lu",c);
}

#define NMAX 15
int n=10;

compteur count[NMAX];

point damier[NMAX][2*(NMAX-1)];

int verbose = 0;

int check(int i,int j)
{
  if ((0 <= i) && (i < n) && 0 <= j && (j < 2*(n-1))) {
    return ((i > 0) || (j > n-2));
  } else {
   return 0;
  }
}

void set_nexts(int i,int j) {
  int k=0;
  point *r = &damier[i][j];

  if (check(i+1,j)) {r->voisins[k++] = &damier[i+1][j];}
  if (check(i,j+1)) {r->voisins[k++] = &damier[i][j+1];}
  if (check(i-1,j)) {r->voisins[k++] = &damier[i-1][j];}
  if (check(i,j-1)) {r->voisins[k++] = &damier[i][j-1];}
  r->voisins[k++] = 0;
}

void init_damier()
{
  int i,j;

  for (i=0;i<n;i++) {
    for (j=0;j<2*(n-1);j++) {
      set_nexts(i,j);
    }
  }
}


void print_sol()
{
  int i,j;
  point *r;

  for (i=n-1;i >= 0;i--) {
    for (j=0;j<2*(n-1);j++) {
      r = &damier[i][j];
      putchar(r->stat==libre?'.':(r->stat==plein?'#':'X'));
    }
    putchar('\n');
  }
  putchar('\n');
}


void search(int k,point *nexts) {

  if (k > 0) {
    point *me,*news,**p;

    for (;nexts;) {
      me = nexts;
      nexts = nexts->next;
      me->stat = plein;
      INCR(count[k]);
      /*      print_sol(); */
      news = nexts;
      for (p=&(me->voisins[0]) ; *p ; p++) {
        if ((*p)->stat == libre) {
          (*p)->stat = voisin;
          (*p)->next = news;
          news = *p;
        }
      }
      search(k-1,news);
      for (; news != nexts; news = news->next) {
        news->stat = libre;
      }
      me->stat = voisin;
    }
  } else {
    for (; nexts ; nexts = nexts->next ) {
      INCR(count[k]);
      nexts->stat = plein; 
      if (verbose) {
        print_sol();
      }
      nexts->stat = voisin;
    }
  }
}


void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-v]* [n]\n",pgm);
  exit(-1);
}


int main(int argc,char **argv)
{
  int i;
  point *p;
  char *arg = NULL ;
  char *pgm = argv[0];

  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') {
        verbose++;
      } else
        usage(pgm);
    } else {
      break;
    }
  }
  if (argc == 1) {
    n = atoi(arg);
  } else if (argc != 0) {
    usage(pgm);
  }
  if (n > NMAX) {
    fprintf(stderr,"argument must be <= %d\n",NMAX);
    exit(2);
  }

  init_damier();

  p=&damier[0][n-2];
  p->next = 0; p->stat = voisin;

  search(n-1,p);

  for (i=0; i<n ; i++) {
    printf("%2d ",i+1);
    print_count(stdout,count[n-i-1]);
    putchar('\n');
  }
  return 0;
}















